/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/** File:           settime.c                                             **/
/** Author:         Marc Mosthav, SIEMENS plc                             **/
/** Created:        Sat Jul 25 10:14:29 BST 1998                          **/
/** Last Modified:  Sat Aug  1 20:16:56 BST 1998                          **/
/**                                                                       **/
/** This implements a SNTP client for setting local time from IPv6 or     **/
/** IPv4 server                                                           **/
/**                                                                       **/
/** PLEASE READ README.ipv6 for copyright information                     **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/in6.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>

#include "sockhelp.h"

/* D1970 contains the number of seconds between 1/1/1900 and 1/1/1970 */
unsigned long D1970=2208988800UL;

/* Names for days of the week     */
char *wday[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

/* IP version only values of 4 or 6 are allowed */
int ipv=0;

/* buffer for addresses */
char addr[48];

/***************************************************************************/
/*                                                                         */
/* Function:   usage                                                       */
/* Parameters: name - name of the program                                  */
/*                                                                         */
/* Print out a usage message                                               */
/*                                                                         */
/***************************************************************************/
void usage(char *name)
{
	fprintf(stderr,"usage: %s [opt]\n\n",name);
	fprintf(stderr,"OPTIONS:\n");
	fprintf(stderr,"\t-4 addr\n");
	fprintf(stderr,"\t\tdefine an IPv4 server to sync with\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"\t\tcan not be used together with -6 or\n\n");
	fprintf(stderr,"\t-6 addr\n");
	fprintf(stderr,"\t\tdefine an IPv6 server to sync with\n");
	fprintf(stderr,"\t\tcan not be used together with -4 or\n\n");
	exit(1);
}

/***************************************************************************/
/*                                                                         */
/* Function:   do_opts                                                     */
/* Parameters: argc: number of arguments                                   */
/*             argv: pointer to arguments                                  */
/*                                                                         */
/* Process programme parameters                                            */
/*                                                                         */
/***************************************************************************/
void do_opts(int argc, char *argv[])
{
	int c;

	while((c=getopt(argc,argv,"4:6:"))!=EOF) {
		switch(c) {
		/* Set IPv6 address to sync from */
		case '6':
				/* print usage if IP version already set */
				if(ipv) usage(argv[0]);
				/* copy address */
				strcpy(addr,optarg);
				/* set IP version */
				ipv=6;
				break;
		/* Set IPv4 address to sync from */
		case '4':
				/* print usage if IP version already set */
				if(ipv) usage(argv[0]);
				/* copy address */
				strcpy(addr,optarg);
				/* set IP version */
				ipv=4;
				break;
		/* unknown option => print usage */
		default:
				usage(argv[0]);
		}
	}
	/* no IP version set => print usage */
	if(ipv==0) usage(argv[0]);
}

/***************************************************************************/
/*                                                                         */
/* Function:   main                                                        */
/* Parameters: argc: number of arguments                                   */
/*             argv: pointer to arguments                                  */
/*                                                                         */
/* main function of the programme                                          */
/*                                                                         */
/***************************************************************************/
int main(argc, argv)
int argc;
char *argv[];
{
	int sock; /* Our socket that we will talk through */
	/* Client and server address info (IPv6) */
	struct sockaddr_in6 client6, server6;
	/* Client and server address info (IPv4) */
	struct sockaddr_in client, server;
	int recvd; /* Number of bytes recieved */
	int structlength; /* Length of sockaddr structure */
	int port; /* Port we will talk on */
	struct in6_addr *iaddr6; /* IPv6 address of server */
	struct in_addr *iaddr; /* IPv4 address of server */
	/* buffers for send and receive */
	char buf[48];
	char buffer[48];
	/* variable to store time */
	struct tm *t;
	/* variables for time calculations */
	unsigned long x, y;
	/* counter */
	int i;

	/* check that only super user uses this programm */
	if(getuid()!=0) {
		fprintf(stderr,"You have to be superuser!!!\n");
		exit(3);
	}

	/* process options */
	do_opts(argc,argv);

	/* get port number */
	port = htons(123);
	if (port == -1) {
		fprintf(stderr,"Can't use port: %s\n","ntp");
		exit(EXIT_FAILURE);
	}

	if(ipv==4) {
		iaddr=atoaddr(addr);
		if (iaddr == NULL) {
			fprintf(stderr,"Invalid network address: %s\n",addr);
			exit(EXIT_FAILURE);
		}
		inet_ntop(AF_INET,iaddr,buffer,47);
		printf("Getting time from %s...\n",buffer);
	} else {
		iaddr6=atoaddr6(addr);
		if (iaddr6 == NULL) {
			fprintf(stderr,"Invalid network address: %s\n",addr);
			exit(EXIT_FAILURE);
		}
		inet_ntop(AF_INET6,iaddr6,buffer,47);
		printf("Getting time from %s...\n",buffer);
	}

	bzero(buf,48);
	buf[0]=0x0b;

	/* open socket */
	if(ipv==4)
		sock = socket(AF_INET, SOCK_DGRAM, 0);
	else
		sock = socket(AF_INET6, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("socket");
		exit(EXIT_FAILURE);
	}

	if(ipv==4) {
		/* Set up the server info */
		memset((char *) &server, 0, sizeof(server));
		server.sin_family = AF_INET;
		server.sin_addr.s_addr = iaddr->s_addr;
		server.sin_port = port;

		/* Set up another structure for our local socket */
		memset((char *) &client, 0, sizeof(client));
		client.sin_family = AF_INET;
		client.sin_addr.s_addr = htonl(INADDR_ANY);
		client.sin_port = htons(0);

		/* bind address to socket*/
		if(bind(sock,(struct sockaddr *)&client,sizeof(client)) < 0) {
			perror("bind");
			exit(EXIT_FAILURE);
		}

		/* send request */
		if (sendto(sock, &buf, sizeof(buf), 0, 
		(struct sockaddr *) &server, sizeof(server)) < 0 ) { 
			perror("sendto");
			exit(EXIT_FAILURE);
		}

		/* Then wait for response */
		structlength = sizeof(client);
		recvd = recvfrom(sock, &buf, sizeof(buf), 0,
		(struct sockaddr *) &client, &structlength);
		if (recvd < 0) {
			perror("recvfrom");
			exit(EXIT_FAILURE);
		}
		if (recvd != sizeof(buf)) {
			fprintf(stderr,"Got back wrong number of bytes!\n");
			exit(EXIT_FAILURE);
		}
	} else {
		/* Set up the server info */
		memset((char *) &server6, 0, sizeof(server6));
		server6.sin6_family = AF_INET6;
		bcopy(iaddr6,&(server6.sin6_addr),16);
		server6.sin6_port = port;

		/* Set up another structure for our local socket */
		memset((char *) &client6, 0, sizeof(client6));
		client6.sin6_family = AF_INET6;
		client6.sin6_addr = in6addr_any;
		client6.sin6_port = htons(0);

		/* bind address to socket*/
		if(bind(sock,(struct sockaddr *)&client6,sizeof(client6))<0) {
			perror("bind");
			exit(EXIT_FAILURE);
		}

		/* send request */
		if (sendto(sock, &buf, sizeof(buf), 0, 
		(struct sockaddr *) &server6, sizeof(server6)) < 0 ) { 
			perror("sendto");
			exit(EXIT_FAILURE);
		}

		/* Then wait for response */
		structlength = sizeof(client6);
		recvd = recvfrom(sock, &buf, sizeof(buf), 0,
		(struct sockaddr *) &client6, &structlength);
		if (recvd < 0) {
			perror("recvfrom");
			exit(EXIT_FAILURE);
		}
		if (recvd != sizeof(buf)) {
			fprintf(stderr,"Got back wrong number of bytes!\n");
			exit(EXIT_FAILURE);
		}
	}


	/* get seconds since 1/1/1900 from data */
	y=0UL;
	for(i=0;i<4;i++) {
		y=y<<8;
		y+=(buf[40+i]&0xff);
	}

	/* calculate seconds since 1/1/1970 */
	x=y-D1970;

	/* set system time */
	stime(&x);

	/* get local time */
	t=localtime(&x);

	/* print time */
	printf("%d:",t->tm_hour);
	if(t->tm_min<10)
		printf("0%d:",t->tm_min);
	else
		printf("%d:",t->tm_min);
	if(t->tm_sec<10)
		printf("0%d",t->tm_sec);
	else
		printf("%d",t->tm_sec);
	printf(", %s ",wday[t->tm_wday]);
	printf("%d/",t->tm_mday);
	printf("%d/",1+t->tm_mon);
	printf("%d\n",1900+t->tm_year);

	return 0;
}
