/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/** File:           sntpcd.c                                              **/
/** Author:         Marc Mosthav, SIEMENS plc                             **/
/** Created:        Sat Jul 25 12:32:18 BST 1998                          **/
/** Last Modified:  Sat Aug  1 19:15:36 BST 1998                          **/
/**                                                                       **/
/** This implements a SNTP client daemon listening to IPv6 multicasts     **/
/**                                                                       **/
/** PLEASE READ README.ipv6 for copyright information                     **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/in6.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>

#include "sockhelp.h"

/* D1970 contains the number of seconds between 1/1/1900 and 1/1/1970 */
unsigned long D1970=2208988800UL;

/* verbose modes are:             */
/*                   0: quiet     */
/*                   1: info      */
/*                   2: loud      */
int verbose=0;

/* Names for days of the week     */
char *wday[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

/* Multicast mode (1=on, 0=off) but I'm not really using it yet */
int multi=1;

/* IPv6 address of multicast group */
char *maddr="FF02::101";

/* IP version only values of 4 or 6 are allowed */
int ipv=0;

/* buffer for addresses */
char addr[48];

/* pointer to reference ID initialised to LOCL for local clock */
char *ref="LOCL";

/* are we a primary server? Default=yes ( 1=yes, 0=no ) */
int primary=1;

/* global pointer to store argv[0] */
char *prg;

/***************************************************************************/
/*                                                                         */
/* Function:   usage                                                       */
/* Parameters: name - name of the program                                  */
/*                                                                         */
/* Print out a usage message                                               */
/*                                                                         */
/***************************************************************************/
void usage(char *name)
{
	fprintf(stderr,"usage: %s [opt] devices\n\n",name);
	fprintf(stderr,"OPTIONS:\n");
	fprintf(stderr,"\t-s scope\n");
	fprintf(stderr,"\t\tscope can be n,l,s,o or g where:\n");
	fprintf(stderr,"\t\tn is for multicast node-local\n");
	fprintf(stderr,"\t\tl is for multicast link-local\n");
	fprintf(stderr,"\t\ts is for multicast site-local\n");
	fprintf(stderr,"\t\to is for multicast organization-local\n");
	fprintf(stderr,"\t\tg is for multicast global\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"\t-4 addr\n");
	fprintf(stderr,"\t\tdefine an IPv4 server to sync with\n");
	fprintf(stderr,"\t\tcan not be used together with -6 or -p\n\n");
	fprintf(stderr,"\t-6 addr\n");
	fprintf(stderr,"\t\tdefine an IPv6 server to sync with\n");
	fprintf(stderr,"\t\tcan not be used together with -4 or -p\n\n");
	fprintf(stderr,"\t-v level\n");
	fprintf(stderr,"\t\tlevel of verbosity where level is:\n");
	fprintf(stderr,"\t\t0: quiet\n\t\t1: informational\n\t\t2: loud\n\n");
	exit(1);
}

/***************************************************************************/
/*                                                                         */
/* Function:   do_opts                                                     */
/* Parameters: argc: number of arguments                                   */
/*             argv: pointer to arguments                                  */
/*                                                                         */
/* Process programme parameters                                            */
/*                                                                         */
/***************************************************************************/
void do_opts(int argc, char *argv[])
{
	int c;

	while((c=getopt(argc,argv,"s:4:6:"))!=EOF) {
		switch(c) {
		/* Set IPv6 address to sync from */
		case '6':
				/* print usage if IP version already set */
				if(ipv) usage(argv[0]);
				/* copy address */
				strcpy(addr,optarg);
				/* set IP version */
				ipv=6;
				/* We are syncing => we're not primary */
				primary=0;
				break;
		/* Set IPv4 address to sync from */
		case '4':
				/* print usage if IP version already set */
				if(ipv) usage(argv[0]);
				/* copy address */
				strcpy(addr,optarg);
				/* set IP version */
				ipv=4;
				/* We are syncing => we're not primary */
				primary=0;
				break;
		/* select multicast scope */
		case 's':
				switch(optarg[0]) {
				/* scope node local */
				case 'n':
						multi=1;
						maddr="FF01::101";
						break;
				/* scope link local */
				case 'l':
						multi=1;
						maddr="FF02::101";
						break;
				/* scope site local */
				case 's':
						multi=1;
						maddr="FF05::101";
						break;
				/* scope organisation local */
				case 'o':
						multi=1;
						maddr="FF08::101";
						break;
				/* scope global */
				case 'g':
						multi=1;
						maddr="FF0E::101";
						break;
				/* unknown scope => print usage */
				default:
						usage(argv[0]);
				}
				break;
		/* unknown option => print usage */
		default:
				usage(argv[0]);
		}
	}
	/* no IP version set => print usage */
	if(ipv==0) usage(argv[0]);
}

/***************************************************************************/
/*                                                                         */
/* Function:   add_multi_devs                                              */
/* Parameters: sock: socket identifier                                     */
/*             devc: number of devices                                     */
/*             dev:  pointer to names of devices                           */
/*                                                                         */
/*                                                                         */
/* function for setting multicast options to every device (integer index)  */
/* for multicast membership on argumented socket                           */
/*                                                                         */
/***************************************************************************/
void add_multi_devs(int sock, int devc, char **dev)
{
        int i;

        struct ipv6_mreq multi;/* multicast structure  
	                          defined in <netinet/in.h>*/

	if(devc<=0) usage(prg);
	if(verbose>1) {
		fprintf(stderr,"->add_multi_devs\n");
	}
        for(i=0;i<devc;i++) {
		if(verbose)
			fprintf(stderr,"adding multi %s on %s\n",maddr,dev[i]);
		/* get index for interface name */
                multi.ipv6mr_ifindex=if_nametoindex(dev[i]);
		/* convert string representation to net format */
                inet_pton(AF_INET6,maddr,&(multi.ipv6mr_multiaddr));
		/* add muticast group membership */
                setsockopt(sock, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP,
			(char *)&multi,sizeof(struct ipv6_mreq));
        }
}

/***************************************************************************/
/*                                                                         */
/* Function:   main                                                        */
/* Parameters: argc: number of arguments                                   */
/*             argv: pointer to arguments                                  */
/*                                                                         */
/* main function of the programme                                          */
/*                                                                         */
/***************************************************************************/
int main(argc, argv)
int argc;
char *argv[];
{
	int sock; /* Our socket that we will talk through */
	/* Client and server address info (IPv6) */
	struct sockaddr_in6 client6, server6;
	/* Client and server address info (IPv4) */
	struct sockaddr_in client, server;
	int recvd; /* Number of bytes recieved */
	int structlength; /* Length of sockaddr structure */
	int port; /* Port we will talk on */
	struct in6_addr *iaddr6; /* IPv6 address of server */
	struct in_addr *iaddr; /* IPv4 address of server */
	/* buffer for send and receive */
	char buf[48];
	/* variable to store time */
	struct tm *t;
	/* variables for time calculations */
	unsigned long x, y;
	/* counter */
	int i;

  
	/* save name of programme in global variable */
	prg=argv[0];

	/* check that only super user uses this programm */
	if(getuid()!=0) {
		fprintf(stderr,"You have to be superuser!!!\n");
		exit(3);
	}

	/* process options */
	do_opts(argc,argv);

	/* get port number */
	port = htons(123);
	if (port == -1) {
		fprintf(stderr,"Can't use port: %s\n","ntp");
		exit(EXIT_FAILURE);
	}

	/* open socket */
	sock = socket(AF_INET6, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("socket");
		exit(EXIT_FAILURE);
	}

	/* join multicast groups */
	add_multi_devs(sock,argc-optind,&argv[optind]);

	/* Set up structure for our local socket */
	memset((char *) &client6, 0, sizeof(client6));
	client6.sin6_family = AF_INET6;
	client6.sin6_flowinfo = 0;
	client6.sin6_addr = in6addr_any;
	client6.sin6_port = port;

	/* bind address to socket*/
	if(bind(sock, (struct sockaddr *)&client6, 
	sizeof(struct sockaddr_in6)) == -1) {
		perror("bind");
		exit(2);
	}

	/* get length of sockaddr structure */
	structlength = sizeof(client6);

	/* do an endless loop - we are a daemon */
	while(1) {
		/* wait for data to receive */
		recvd = recvfrom(sock, &buf, sizeof(buf), 0,
			(struct sockaddr *) &client6, &structlength);
		if (recvd < 0) {
			perror("recvfrom");
			continue;
		}
		if (recvd != sizeof(buf)) {
			fprintf(stderr,"Got back wrong number of bytes!\n");
			continue;
		}

		/* get seconds since 1/1/1900 from data */
		y=0UL;
		for(i=0;i<4;i++) {
			y=y<<8;
			y+=(buf[40+i]&0xff);
		}

		/* calculate seconds since 1/1/1970 */
		x=y-D1970;

		/* set system time */
		stime(&x);

		/* if verbose print the time */
		if(verbose) {
			/* get local time */
			t=localtime(&x);
			printf("%d:",t->tm_hour);
			if(t->tm_min<10)
				printf("0%d:",t->tm_min);
			else
				printf("%d:",t->tm_min);
			if(t->tm_sec<10)
				printf("0%d",t->tm_sec);
			else
				printf("%d",t->tm_sec);
			printf(", %s ",wday[t->tm_wday]);
			printf("%d/",t->tm_mday);
			printf("%d/",1+t->tm_mon);
			printf("%d\n",1900+t->tm_year);
		}
	}
	/* Really we should never get here ;-) */
	return 0;
}
