/***************************************************************************/
/***************************************************************************/
/**                                                                       **/
/** File:           sntpsd.c                                              **/
/** Author:         Marc Mosthav, SIEMENS plc                             **/
/** Created:        Sat Jul 25 18:17:19 BST 1998                          **/
/** Last Modified:  Sun Aug  2 10:49:58 BST 1998                          **/
/**                                                                       **/
/** This implements a SNTP server daemon sending IPv6 multicasts          **/
/**                                                                       **/
/** PLEASE READ README.ipv6 for copyright information                     **/
/**                                                                       **/
/***************************************************************************/
/***************************************************************************/

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/in6.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include "sockhelp.h"
#include <time.h>

/* D1970 contains the number of seconds between 1/1/1900 and 1/1/1970 */
unsigned long D1970=2208988800UL;

/* Names for days of the week     */
char *wday[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

/* Multicast mode (1=on, 0=off) but I'm not really using it yet */
int multi=0;

/* IPv6 address of multicast group */
char *maddr="FF02::101";

/* IP version only values of 4 or 6 are allowed */
int ipv=0;

/* buffer for addresses and data packets */
char addr[48];
char gbuffer[48];
char sbuf[48];

/* pointer to reference ID initialised to LOCL for local clock */
char *ref;

/* are we a primary server? Default=yes ( 1=yes, 0=no ) */
int primary=0;

/* intervall */
int interv=10;

/* precision */
int prec=-6;


/***************************************************************************/
/*                                                                         */
/* Function:   usage                                                       */
/* Parameters: name - name of the program                                  */
/*                                                                         */
/* Print out a usage message                                               */
/*                                                                         */
/***************************************************************************/
void usage(char *name)
{
	fprintf(stderr,"usage: %s [opt]\n\n",name);
	fprintf(stderr,"OPTIONS:\n");
	fprintf(stderr,"\t-s scope\n");
	fprintf(stderr,"\t\tscope can be n,l,s,o or g where:\n");
	fprintf(stderr,"\t\tn is for multicast node-local\n");
	fprintf(stderr,"\t\tl is for multicast link-local\n");
	fprintf(stderr,"\t\ts is for multicast site-local\n");
	fprintf(stderr,"\t\to is for multicast organization-local\n");
	fprintf(stderr,"\t\tg is for multicast global\n");
	fprintf(stderr,"\n");
	fprintf(stderr,"\t-4 addr\n");
	fprintf(stderr,"\t\tdefine an IPv4 server to sync with\n");
	fprintf(stderr,"\t\tcan not be used together with -6 or -p\n\n");
	fprintf(stderr,"\t-6 addr\n");
	fprintf(stderr,"\t\tdefine an IPv6 server to sync with\n");
	fprintf(stderr,"\t\tcan not be used together with -4 or -p\n\n");
	fprintf(stderr,"\t-p source\n");
	fprintf(stderr,"\t\tdefines the server to be primary. The source\n");
	fprintf(stderr,"\t\tshould be one of the following. If non of the\n");
	fprintf(stderr,"\t\tfollowing applies any string can be used\n\n");
	fprintf(stderr,"\t\tSource   External Reference\n");
	fprintf(stderr,"\t\t------------------------------------------------\n");
	fprintf(stderr,"\t\tLOCL     uncalibrated local clock used as a primary\n");

	fprintf(stderr,"\t\t         reference for a subnet without external\n");
	fprintf(stderr,"\t\t         means of synchronization\n");
	fprintf(stderr,"\t\tPPS      atomic clock or other pulse-per-second\n");
	fprintf(stderr,"\t\t         source individually calibrated to national\n");
	fprintf(stderr,"\t\t         standards\n");
	fprintf(stderr,"\t\tACTS     NIST dialup modem service\n");
	fprintf(stderr,"\t\tUSNO     USNO modem service\n");
	fprintf(stderr,"\t\tPTB      PTB (Germany) modem service\n");
	fprintf(stderr,"\t\tTDF      Allouis (France) Radio 164 kHz\n");
	fprintf(stderr,"\t\tDCF      Mainflingen (Germany) Radio 77.5 kHz\n");
	fprintf(stderr,"\t\tMSF      Rugby (UK) Radio 60 kHz\n");
	fprintf(stderr,"\t\tWWV      Ft. Collins (US) Radio 2.5,5,10,15,20 MHz\n");
	fprintf(stderr,"\t\tWWVB     Boulder (US) Radio 60 kHz\n");
	fprintf(stderr,"\t\tWWVH     Kaui Hawaii (US) Radio 2.5, 5, 10, 15 MHz\n");
	fprintf(stderr,"\t\tCHU      Ottawa (Canada) Radio 3330, 7335, 14670 kHz\n");
	fprintf(stderr,"\t\tLORC     LORAN-C radionavigation system\n");
	fprintf(stderr,"\t\tOMEG     OMEGA radionavigation system\n");
	fprintf(stderr,"\t\tGPS      Global Positioning Service\n");
	fprintf(stderr,"\t\tGOES     Geostationary Orbit Environment Satellite\n\n");
	fprintf(stderr,"\t-i interval\n");
	fprintf(stderr,"\t\tinterval is an eight-bit signed integer indicating\n");
	fprintf(stderr,"\t\tthe maximum interval between successive messages,\n");
	fprintf(stderr,"\t\tin seconds to the nearest power of two. The values\n");
	fprintf(stderr,"\t\that can appear in this field presently range from\n");
	fprintf(stderr,"\t\t4 (16 s) to 14 (16284 s)\n\n");
	fprintf(stderr,"\t-c precision\n");
	fprintf(stderr,"\t\tprecision is an eight-bit signed integer indicating\n");
	fprintf(stderr,"\t\tthe precision of the local clock, in seconds to\n");
	fprintf(stderr,"\t\tthe nearest power of two.  The values that normally\n");
	fprintf(stderr,"\t\tappear in this field range from -6 for mains-\n");
	fprintf(stderr,"\t\tfrequency clocks to -20 for microsecond clocks\n");
	fprintf(stderr,"\t\tfound in some workstations\n");
	exit(1);
}

/***************************************************************************/
/*                                                                         */
/* Function:   do_opts                                                     */
/* Parameters: argc: number of arguments                                   */
/*             argv: pointer to arguments                                  */
/*                                                                         */
/* Process programme parameters                                            */
/*                                                                         */
/***************************************************************************/
void do_opts(int argc, char *argv[])
{
	int c;

	while((c=getopt(argc,argv,"s:p:4:6:i:c:"))!=EOF) {
		switch(c) {
		/* Set intervall */
		case 'i':
				interv=atoi(optarg);
				break;
		/* Set precision */
		case 'c':
				prec=atoi(optarg);
				break;
		/* Set primary */
		case 'p':
				/* print usage if IP version already set */
				if(ipv) usage(argv[0]);
				/* copy refernce ID */
				bzero(ref,4);
				strncpy(ref,optarg,4);
				primary=1;
				break;
		/* Set IPv6 address to sync from */
		case '6':
				/* print usage if IP version already set */
				if(ipv) usage(argv[0]);
				/* copy address */
				strcpy(addr,optarg);
				/* set IP version */
				ipv=6;
				/* We are syncing => we're not primary */
				primary=0;
				break;
		/* Set IPv4 address to sync from */
		case '4':
				/* print usage if IP version already set */
				if(ipv) usage(argv[0]);
				/* copy address */
				strcpy(addr,optarg);
				/* set IP version */
				ipv=4;
				/* We are syncing => we're not primary */
				primary=0;
				break;
		/* select multicast scope */
		case 's':
				switch(optarg[0]) {
				/* scope node local */
				case 'n':
						multi=1;
						maddr="FF01::101";
						break;
				/* scope link local */
				case 'l':
						multi=1;
						maddr="FF02::101";
						break;
				/* scope site local */
				case 's':
						multi=1;
						maddr="FF05::101";
						break;
				/* scope organisation local */
				case 'o':
						multi=1;
						maddr="FF08::101";
						break;
				/* scope global */
				case 'g':
						multi=1;
						maddr="FF0E::101";
						break;
				/* unknown scope => print usage */
				default:
						usage(argv[0]);
				}
				break;
		/* unknown option => print usage */
		default:
				usage(argv[0]);
		}
	}
	/* no IP version set and not primary => print usage */
	if((ipv==0)&&(!primary)) usage(argv[0]);
}

/***************************************************************************/
/*                                                                         */
/* Function:   msend                                                       */
/*                                                                         */
/* Send SNTP packets via multicast                                         */
/*                                                                         */
/***************************************************************************/
void msend(void)
{
	/* Server address info (IPv6) */
	struct sockaddr_in6 server6;
	int sock; /* socket we want to talk through */
	/* variable for time calculations */
	unsigned long t;
	/* counter */
	int i;

	/* fill in address structure */
	memset((char *) &server6, 0, sizeof(server6));
	server6.sin6_family = AF_INET6;
	server6.sin6_port = htons(123);
	inet_pton(AF_INET6,maddr,&(server6.sin6_addr));
	sock = socket(AF_INET6, SOCK_DGRAM, 0);

	/* get time (seconds since 1/1/1970) */
	t=(unsigned long)time(NULL);

	/* calculate seconds since 1/1/1900 */
	t+=D1970;
	for(i=0;i<4;i++) {
		gbuffer[43-i]=t&0xff;
		t=t>>8;
	}

	/* send multicast */
	if (sendto(sock, gbuffer, sizeof(gbuffer), 0, 
	(struct sockaddr *) &server6, sizeof(server6)) < 0 ) { 
		perror("sendto");
		exit(EXIT_FAILURE);
	}

	/* close socket */
	if(sock) close(sock);
}

/***************************************************************************/
/*                                                                         */
/* Function:   do_sync                                                     */
/*                                                                         */
/* Syncronise with primary server                                          */
/*                                                                         */
/***************************************************************************/
void do_sync(void)
{
	int sock; /* Our socket that we will talk through */
	/* Client and server address info (IPv6) */
	struct sockaddr_in6 client6, server6;
	/* Client and server address info (IPv4) */
	struct sockaddr_in client, server;
	int recvd; /* Number of bytes recieved */
	int structlength; /* Length of sockaddr structure */
	int port; /* Port we will talk on */
	struct in6_addr *iaddr6; /* Internet address of server */
	struct in_addr *iaddr; /* Internet address of server */
	char buf[48]; /* buffer for receiving data */
	/* variable to store time */
	struct tm *t;
	/* variables for time calculations */
	unsigned long x, y;
	/* counter */
	int i;

	/* don't sync if we are primary */
	if(primary) return;

	/* get port number */
	port = htons(123);

	/* get address of server to sync with */
	if(ipv==4) {
		iaddr=atoaddr(addr);
		if (iaddr == NULL) {
			fprintf(stderr,"Invalid network address: %s\n",addr);
			exit(EXIT_FAILURE);
		}
	} else {
		iaddr6=atoaddr6(addr);
		if (iaddr6 == NULL) {
			fprintf(stderr,"Invalid network address: %s\n",addr);
			exit(EXIT_FAILURE);
		}
	}

	/* open socket */
	if(ipv==4)
		sock = socket(AF_INET, SOCK_DGRAM, 0);
	else
		sock = socket(AF_INET6, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("socket");
		return;
	}

	if(ipv==4) {
		/* Set up the server info */
		memset((char *) &server, 0, sizeof(server));
		server.sin_family = AF_INET;
		server.sin_addr.s_addr = iaddr->s_addr;
		server.sin_port = port;

		/* Set up another structure for our local socket */
		memset((char *) &client, 0, sizeof(client));
		client.sin_family = AF_INET;
		client.sin_addr.s_addr = htonl(INADDR_ANY);
		client.sin_port = htons(0);
		if(bind(sock,(struct sockaddr *)&client,sizeof(client)) < 0) {
			perror("bind");
			return;
		}

		/* send request */
		if (sendto(sock, sbuf, sizeof(sbuf), 0, 
		(struct sockaddr *) &server, sizeof(server)) < 0 ) { 
			perror("sendto");
			return;
		}

		/* wait for answer */
		structlength = sizeof(client);
		recvd = recvfrom(sock, buf, sizeof(buf), 0,
		(struct sockaddr *) &client, &structlength);
		if (recvd < 0) {
			perror("recvfrom");
			return;
		}
		if (recvd != sizeof(buf)) {
			fprintf(stderr,"Got back wrong number of bytes!\n");
			return;
		}
	} else {
		/* Set up the server info */
		memset((char *) &server, 0, sizeof(server));
		server6.sin6_family = AF_INET6;
		server6.sin6_addr = *iaddr6;
		server6.sin6_port = port;

		/* Set up another structure for our local socket */
		memset((char *) &client6, 0, sizeof(client6));
		client6.sin6_family = AF_INET6;
		client6.sin6_addr = in6addr_any;
		client6.sin6_port = htons(0);
		if(bind(sock,(struct sockaddr *)&client6,sizeof(client6))<0) {
			perror("bind");
			return;
		}

		/* send request */
		if (sendto(sock, sbuf, sizeof(sbuf), 0, 
		(struct sockaddr *) &server6, sizeof(server6)) < 0 ) { 
			perror("sendto");
			return;
		}

		/* wait for answer */
		structlength = sizeof(client6);
		recvd = recvfrom(sock, buf, sizeof(buf), 0,
		(struct sockaddr *) &client6, &structlength);
		if (recvd < 0) {
			perror("recvfrom");
			return;
		}
		if (recvd != sizeof(buf)) {
			fprintf(stderr,"Got back wrong number of bytes!\n");
			exit(EXIT_FAILURE);
		}
	}

	/* get seconds since 1/1/1900 from data */
	y=0UL;
	for(i=0;i<4;i++) {
		y=y<<8;
		y+=(buf[40+i]&0xff);
	}

	/* calculate seconds since 1/1/1970 */
	x=y-D1970;

	/* set system time */
	stime(&x);

	/* set reference id */
	bcopy(&buf[44],ref,4);

	/* close socket */
	if(sock) close(sock);
}

/***************************************************************************/
/*                                                                         */
/* Function:   h_alarm                                                     */
/*                                                                         */
/* Process alarm signal                                                    */
/*                                                                         */
/***************************************************************************/
void h_alarm(void)
{
	/* flip is used to switch between syncing and sending multicasts */
	static int flip=0;

	/* set signal handling function for alarm signal */
	(void)signal(SIGALRM,(void *)h_alarm);

	if(flip) { /* send multicast */
		msend();
		/* set alarm to 37 seconds */
		alarm(37);
	} else {   /* sync with primary server */
		do_sync();
		/* set alarm to 46 seconds */
		alarm(46);
	}

	/* change state of flip variable */
	flip=!flip;
}

/***************************************************************************/
/*                                                                         */
/* Function:   main                                                        */
/* Parameters: argc: number of arguments                                   */
/*             argv: pointer to arguments                                  */
/*                                                                         */
/* main function of the programme                                          */
/*                                                                         */
/***************************************************************************/
int main(argc, argv)
int argc;
char *argv[];
{
	int sock;  /* Sock we will talk on */
	/* Address information for client and server */
	struct sockaddr_in6 client, server;
	int recvd; /* Number of bytes recieved */
	int structlength; /* Length of sockaddr structure */
	/* buffer for send/receive */
	char buffer[48];
	/* variable for time calculations */
	unsigned long t;
	/* counter */
	int i;

	/* check that only super user uses this programm */
	if(getuid()!=0) {
		fprintf(stderr,"You have to be superuser!!!\n");
		exit(3);
	}

	/* build multicast packet */
	bzero(gbuffer,48);
	ref=&gbuffer[12];
	bcopy("LOCL",ref,4);
	do_opts(argc,argv);
	gbuffer[0]=0x25;
	gbuffer[1]=2-primary;
	gbuffer[2]=interv;
	gbuffer[3]=prec;

	/* build packet to send for syncing */
	bzero(sbuf,48);
	sbuf[0]=0x0b;

	/* sync with primary server */
	do_sync();

	/* send out multicast */
	msend();

	/* fill in address structure */
	server.sin6_family=AF_INET6;
	server.sin6_addr=in6addr_any;
	server.sin6_port=htons(123);

	/* open socket */
	sock = socket(AF_INET6, SOCK_DGRAM, 0);

	/* bind address to socket*/
	if (bind(sock, (struct sockaddr *) &server, sizeof(server)) < 0) {
		perror("bind");
		exit(4);
	}

	/* set signal handling function for alarm signal */
	(void)signal(SIGALRM,(void *)h_alarm);

	/* set alarm to 43 seconds */
	alarm(43);

	/* get length of sockaddr structure */
	structlength = sizeof(client);

	/* endless loop - we are a daemon ;-) */
	while(1) {
		/* wait to receive request */
		recvd = recvfrom(sock, buffer, sizeof(buffer), 0, 
		(struct sockaddr *) &client, &structlength);
		if (recvd < 0) {
			perror("recvfrom");
			continue;
		}

		/* set reverence ID */
		bcopy(ref,&buffer[12],4);

		/* copy originating timestamp */
		bcopy(&buffer[40],&buffer[24],8);
		bzero(&buffer[44],4);

		/* get time (seconds since 1/1/1970) */
		t=(unsigned long)time(NULL);

		/* calculate seconds since 1/1/1900 */
		t+=D1970;
		for(i=0;i<4;i++) {
			buffer[43-i]=t&0xff;
			t=t>>8;
		}

		/* send answer */
		if (sendto(sock, buffer, sizeof(buffer), 0,
		(struct sockaddr *) &client, structlength) < 0) {
			perror("sendto");
			continue;
		}
	}

	/* Really we should never get here ;-) */
	return 0;
}
